@@ -1,4 +1,24 @@
-/* -*- Mode: C; c-basic-offset: 4 -*- */
+/* -*- Mode: C; c-basic-offset: 4 -*-
+ * pygtk- Python bindings for the GTK toolkit.
+ * Copyright (C) 1998-2003  James Henstridge
+ *
+ *   gobjectmodule.c: wrapper for the gobject library.
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * USA
+ */
 
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
@@ -10,6 +30,15 @@ static PyObject *gerror_exc = NULL;
 
 /* -------------- GDK threading hooks ---------------------------- */
 
+/**
+ * pyg_set_thread_block_funcs:
+ * @block_threads_func: a function to block Python threads.
+ * @unblock_threads_func: a function to unblock Python threads.
+ *
+ * an interface to allow pygtk to add hooks to handle threading
+ * similar to the old PyGTK 0.6.x releases.  May not work quite right
+ * anymore.
+ */
 static void
 pyg_set_thread_block_funcs (PyGThreadBlockFunc block_threads_func,
 			    PyGThreadBlockFunc unblock_threads_func)
@@ -26,6 +55,13 @@ object_free(PyObject *op)
     PyObject_FREE(op);
 }
 
+/**
+ * pyg_destroy_notify:
+ * @user_data: a PyObject pointer.
+ *
+ * A function that can be used as a GDestroyNotify callback that will
+ * call Py_DECREF on the data.
+ */
 void
 pyg_destroy_notify(gpointer user_data)
 {
@@ -133,6 +169,14 @@ PyTypeObject PyGParamSpec_Type = {
     NULL
 };
 
+/**
+ * pyg_param_spec_new:
+ * @pspec: a GParamSpec.
+ *
+ * Creates a wrapper for a GParamSpec.
+ *
+ * Returns: the GParamSpec wrapper.
+ */
 PyObject *
 pyg_param_spec_new(GParamSpec *pspec)
 {
@@ -231,9 +275,20 @@ static PyTypeObject PyGInterface_Type = {
     (PyObject *)0,			/* tp_bases */
 };
 
+/**
+ * pyg_register_interface:
+ * @dict: a module dictionary.
+ * @class_name: the class name for the wrapper class.
+ * @gtype: the GType of the interface.
+ * @type: the wrapper class for the interface.
+ *
+ * Registers a Python class as the wrapper for a GInterface.  As a
+ * convenience it will also place a reference to the wrapper class in
+ * the provided module dictionary.
+ */
 static void
 pyg_register_interface(PyObject *dict, const gchar *class_name,
-		       GType gtype, PyTypeObject *type)
+                       GType gtype, PyTypeObject *type)
 {
     PyObject *o;
 
@@ -241,19 +296,260 @@ pyg_register_interface(PyObject *dict, const gchar *class_name,
     type->tp_base = &PyGInterface_Type;
 
     if (PyType_Ready(type) < 0) {
-	g_warning("could not ready `%s'", type->tp_name);
-	return;
+        g_warning("could not ready `%s'", type->tp_name);
+        return;
     }
 
     if (gtype) {
-	o = pyg_type_wrapper_new(gtype);
-	PyDict_SetItemString(type->tp_dict, "__gtype__", o);
-	Py_DECREF(o);
+        o = pyg_type_wrapper_new(gtype);
+        PyDict_SetItemString(type->tp_dict, "__gtype__", o);
+        Py_DECREF(o);
     }
 
     PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
 }
 
+/* -------------- GMainContext objects ---------------------------- */
+
+static int
+pyg_main_context_new(PyGMainContext *self)
+{
+    self->context = g_main_context_new();
+    return 0;
+}
+
+static int
+pyg_main_context_compare(PyGMainContext *self, PyGMainContext *v)
+{
+    if (self->context == v->context) return 0;
+    if (self->context > v->context) return -1;
+    return 1;
+}
+
+static PyObject *
+pyg_main_context_default (PyObject *unused)
+{
+    PyGMainContext *self;
+
+    self = (PyGMainContext *)PyObject_NEW(PyGMainContext,
+					  &PyGMainContext_Type);
+    if (self == NULL)
+	return NULL;
+
+    self->context = g_main_context_default();
+    return (PyObject *)self;
+
+}
+
+static PyObject *
+_wrap_g_main_context_iteration (PyGMainContext *self, PyObject *args)
+{
+    PyObject *py_ret;
+    gboolean may_block = TRUE;
+    
+    if (!PyArg_ParseTuple(args, "|b:GMainContext.iteration",
+			  &may_block))
+	return NULL;
+	
+    py_ret = g_main_context_iteration(self->context, may_block)
+	? Py_True : Py_False;
+    
+    Py_INCREF(py_ret);
+    return py_ret;
+}
+
+static PyObject *
+_wrap_g_main_context_pending (PyGMainContext *self)
+{
+    PyObject *py_ret;
+
+    py_ret = g_main_context_pending(self->context) ? Py_True : Py_False;
+    
+    Py_INCREF(py_ret);
+    return py_ret;
+}
+
+static PyMethodDef _PyGMainContext_methods[] = {
+    { "iteration", (PyCFunction)_wrap_g_main_context_iteration, METH_VARARGS },
+    { "pending", (PyCFunction)_wrap_g_main_context_pending, METH_NOARGS },
+    { NULL, NULL, 0 }
+};
+
+PyTypeObject PyGMainContext_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,			
+    "gobject.MainContext",	
+    sizeof(PyGMainContext),	
+    0,			
+    /* methods */
+    (destructor)0,	
+    (printfunc)0,	
+    (getattrfunc)0,	
+    (setattrfunc)0,	
+    (cmpfunc)pyg_main_context_compare,		
+    (reprfunc)0,	
+    0,			
+    0,		
+    0,		
+    (hashfunc)0,		
+    (ternaryfunc)0,		
+    (reprfunc)0,		
+    (getattrofunc)0,		
+    (setattrofunc)0,		
+    0,				
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, 
+    NULL,
+    (traverseproc)0,		
+    (inquiry)0,
+    (richcmpfunc)0,	
+    0,             
+    (getiterfunc)0,
+    (iternextfunc)0,
+    _PyGMainContext_methods,
+    0,				
+    0,		       	
+    NULL,		
+    NULL,		
+    (descrgetfunc)0,	
+    (descrsetfunc)0,	
+    0,                 
+    (initproc)pyg_main_context_new,
+};
+
+/* -------------- GMainLoop objects ---------------------------- */
+
+static int
+pyg_main_loop_new(PyGMainLoop *self, PyObject *args, PyObject *kwargs)
+{
+
+    static char *kwlist[] = { "context", "is_running", NULL };
+    PyObject *py_context = Py_None;
+    int is_running;
+    GMainContext *context;
+    
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "|Ob:GMainLoop.__init__",
+				     kwlist, &py_context, &is_running))
+        return -1;
+
+    if (!PyObject_TypeCheck(py_context, &PyGMainContext_Type) &&
+	py_context != Py_None) {
+	PyErr_SetString(PyExc_TypeError,
+			"context must be a gobject.GMainContext or None");
+	return -1;
+    }
+
+    if (py_context != Py_None) {
+	context = ((PyGMainContext*)py_context)->context;
+    } else {
+	context = NULL;
+    }
+
+    self->loop = g_main_loop_new(context, is_running);
+    return 0;
+}
+
+static int
+pyg_main_loop_compare(PyGMainLoop *self, PyGMainLoop *v)
+{
+    if (self->loop == v->loop) return 0;
+    if (self->loop > v->loop) return -1;
+    return 1;
+}
+
+static PyObject *
+_wrap_g_main_loop_get_context (PyGMainLoop *loop)
+{
+    PyGMainContext *self;
+
+    self = (PyGMainContext *)PyObject_NEW(PyGMainContext,
+					  &PyGMainContext_Type);
+    
+    self->context = g_main_loop_get_context(loop->loop);
+    
+    if (self->context == NULL)
+	return NULL;
+
+    return (PyObject *)self;
+}
+
+static PyObject *
+_wrap_g_main_loop_is_running (PyGMainLoop *self)
+{
+    PyObject *py_ret;
+    
+    py_ret = g_main_loop_is_running(self->loop) ? Py_True : Py_False;
+    Py_INCREF(py_ret);
+    return py_ret;
+    
+}
+
+static PyObject *
+_wrap_g_main_loop_quit (PyGMainLoop *self)
+{
+    g_main_loop_quit(self->loop);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *
+_wrap_g_main_loop_run (PyGMainLoop *self)
+{
+    g_main_loop_run(self->loop);
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyMethodDef _PyGMainLoop_methods[] = {
+    { "get_context", (PyCFunction)_wrap_g_main_loop_get_context, METH_NOARGS },
+    { "is_running", (PyCFunction)_wrap_g_main_loop_is_running, METH_NOARGS },
+    { "quit", (PyCFunction)_wrap_g_main_loop_quit, METH_NOARGS },
+    { "run", (PyCFunction)_wrap_g_main_loop_run, METH_NOARGS },
+    { NULL, NULL, 0 }
+};
+
+PyTypeObject PyGMainLoop_Type = {
+    PyObject_HEAD_INIT(NULL)
+    0,			
+    "gobject.MainLoop",	
+    sizeof(PyGMainLoop),	
+    0,			
+    /* methods */
+    (destructor)0,	
+    (printfunc)0,	
+    (getattrfunc)0,	
+    (setattrfunc)0,	
+    (cmpfunc)pyg_main_loop_compare,
+    (reprfunc)0,	
+    0,			
+    0,		
+    0,		
+    (hashfunc)0,		
+    (ternaryfunc)0,		
+    (reprfunc)0,		
+    (getattrofunc)0,		
+    (setattrofunc)0,		
+    0,				
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, 
+    NULL,
+    (traverseproc)0,		
+    (inquiry)0,
+    (richcmpfunc)0,	
+    0,             
+    (getiterfunc)0,
+    (iternextfunc)0,
+    _PyGMainLoop_methods,
+    0,				
+    0,		       	
+    NULL,		
+    NULL,		
+    (descrgetfunc)0,	
+    (descrsetfunc)0,	
+    0,                 
+    (initproc)pyg_main_loop_new,
+};
 
 /* ---------------- gobject module functions -------------------- */
 
@@ -381,22 +677,31 @@ pyg_object_set_property (GObject *object, guint property_id,
     PyObject *object_wrapper, *retval;
     PyObject *py_pspec, *py_value;
 
+    pyg_block_threads();
+
     object_wrapper = pygobject_new(object);
-    g_return_if_fail(object_wrapper != NULL);
+
+    if (object_wrapper == NULL) {
+	pyg_unblock_threads();
+	g_return_if_fail(object_wrapper != NULL);
+    }
 
     py_pspec = pyg_param_spec_new(pspec);
     py_value = pyg_value_as_pyobject (value, TRUE);
+
     retval = PyObject_CallMethod(object_wrapper, "do_set_property",
 				 "OO", py_pspec, py_value);
     if (retval) {
 	Py_DECREF(retval);
     } else {
 	PyErr_Print();
-	PyErr_Clear();
     }
+
     Py_DECREF(object_wrapper);
     Py_DECREF(py_pspec);
     Py_DECREF(py_value);
+
+    pyg_unblock_threads();
 }
 
 static void
@@ -406,19 +711,26 @@ pyg_object_get_property (GObject *object, guint property_id,
     PyObject *object_wrapper, *retval;
     PyObject *py_pspec;
 
+    pyg_block_threads();
+
     object_wrapper = pygobject_new(object);
-    g_return_if_fail(object_wrapper != NULL);
+
+    if (object_wrapper == NULL) {
+	pyg_unblock_threads();
+	g_return_if_fail(object_wrapper != NULL);
+    }
 
     py_pspec = pyg_param_spec_new(pspec);
     retval = PyObject_CallMethod(object_wrapper, "do_get_property",
 				 "O", py_pspec);
     if (retval == NULL || pyg_value_from_pyobject(value, retval) < 0) {
 	PyErr_Print();
-	PyErr_Clear();
     }
     Py_DECREF(object_wrapper);
     Py_DECREF(py_pspec);
     Py_XDECREF(retval);
+    
+    pyg_unblock_threads();        
 }
 
 static void
@@ -1192,7 +1504,8 @@ get_handler_priority(gint *priority, PyObject *kwargs)
 			"keyword argument name is not a string");
 	return -1;
     }
-    if (!strcmp(PyString_AsString(key), "priority")) {
+
+    if (strcmp(PyString_AsString(key), "priority") != 0) {
 	PyErr_SetString(PyExc_TypeError,
 			"only 'priority' keyword argument accepted");
 	return -1;
@@ -1335,6 +1648,7 @@ iowatch_marshal(GIOChannel *source, GIOCondition condition, gpointer user_data)
 	res = PyObject_IsTrue(ret);
 	Py_DECREF(ret);
     }
+
     pyg_unblock_threads();
 
     return res;
@@ -1351,7 +1665,7 @@ pyg_io_add_watch(PyObject *self, PyObject *args, PyObject *kwargs)
     len = PyTuple_Size(args);
     if (len < 3) {
 	PyErr_SetString(PyExc_TypeError,
-			"timeout_add requires at least 3 args");
+			"io_add_watch requires at least 3 args");
 	return NULL;
     }
     first = PySequence_GetSlice(args, 0, 3);
@@ -1417,12 +1731,26 @@ static PyMethodDef pygobject_functions[] = {
     { "timeout_add", (PyCFunction)pyg_timeout_add, METH_VARARGS|METH_KEYWORDS },
     { "io_add_watch", (PyCFunction)pyg_io_add_watch, METH_VARARGS|METH_KEYWORDS },
     { "source_remove", pyg_source_remove, METH_VARARGS },
+    { "main_context_default", (PyCFunction)pyg_main_context_default, METH_NOARGS },
     { NULL, NULL, 0 }
 };
 
 
 /* ----------------- Constant extraction ------------------------ */
 
+/**
+ * pyg_constant_strip_prefix:
+ * @name: the constant name.
+ * @strip_prefix: the prefix to strip.
+ *
+ * Advances the pointer @name by strlen(@strip_prefix) characters.  If
+ * the resulting name does not start with a letter or underscore, the
+ * @name pointer will be rewound.  This is to ensure that the
+ * resulting name is a valid identifier.  Hence the returned string is
+ * a pointer into the string @name.
+ *
+ * Returns: the stripped constant name.
+ */
 static char *
 pyg_constant_strip_prefix(gchar *name, const gchar *strip_prefix)
 {
@@ -1441,6 +1769,15 @@ pyg_constant_strip_prefix(gchar *name, const gchar *strip_prefix)
     return name;
 }
 
+/**
+ * pyg_enum_add_constants:
+ * @module: a Python module
+ * @enum_type: the GType of the enumeration.
+ * @strip_prefix: the prefix to strip from the constant names.
+ *
+ * Adds constants to the given Python module for each value name of
+ * the enumeration.  A prefix will be stripped from each enum name.
+ */
 static void
 pyg_enum_add_constants(PyObject *module, GType enum_type,
 		       const gchar *strip_prefix)
@@ -1469,6 +1806,15 @@ pyg_enum_add_constants(PyObject *module, GType enum_type,
     g_type_class_unref(eclass);
 }
 
+/**
+ * pyg_flags_add_constants:
+ * @module: a Python module
+ * @flags_type: the GType of the flags type.
+ * @strip_prefix: the prefix to strip from the constant names.
+ *
+ * Adds constants to the given Python module for each value name of
+ * the flags set.  A prefix will be stripped from each flag name.
+ */
 static void
 pyg_flags_add_constants(PyObject *module, GType flags_type,
 			const gchar *strip_prefix)
@@ -1497,6 +1843,16 @@ pyg_flags_add_constants(PyObject *module, GType flags_type,
     g_type_class_unref(fclass);
 }
 
+/**
+ * pyg_error_check:
+ * @error: a pointer to the GError.
+ *
+ * Checks to see if the GError has been set.  If the error has been
+ * set, then the gobject.GError Python exception will be raised, and
+ * the GError cleared.
+ *
+ * Returns: True if an error was set.
+ */
 static gboolean
 pyg_error_check(GError **error)
 {
@@ -1505,7 +1861,9 @@ pyg_error_check(GError **error)
     if (*error != NULL) {
 	PyObject *exc_instance;
 	PyObject *d;
-
+	
+	pyg_block_threads();
+	
 	exc_instance = PyObject_CallFunction(gerror_exc, "z",
 					     (*error)->message);
 	PyObject_SetAttrString(exc_instance, "domain",
@@ -1525,6 +1883,9 @@ pyg_error_check(GError **error)
 	PyErr_SetObject(gerror_exc, exc_instance);
 	Py_DECREF(exc_instance);
 	g_clear_error(error);
+	
+	pyg_unblock_threads();
+	
 	return TRUE;
     }
     return FALSE;
@@ -1629,6 +1990,20 @@ initgobject(void)
 			 o=pyg_type_wrapper_new(G_TYPE_BOXED));
     Py_DECREF(o);
 
+    PyGMainLoop_Type.ob_type = &PyType_Type;
+    PyGMainLoop_Type.tp_alloc = PyType_GenericAlloc;
+    PyGMainLoop_Type.tp_new = PyType_GenericNew;
+    if (PyType_Ready(&PyGMainLoop_Type))
+	return;
+    PyDict_SetItemString(d, "MainLoop", (PyObject *)&PyGMainLoop_Type);
+    
+    PyGMainContext_Type.ob_type = &PyType_Type;
+    PyGMainContext_Type.tp_alloc = PyType_GenericAlloc;
+    PyGMainContext_Type.tp_new = PyType_GenericNew;
+    if (PyType_Ready(&PyGMainContext_Type))
+	return;
+    PyDict_SetItemString(d, "MainContext", (PyObject *)&PyGMainContext_Type);
+
     PyGPointer_Type.ob_type = &PyType_Type;
     PyGPointer_Type.tp_alloc = PyType_GenericAlloc;
     PyGPointer_Type.tp_new = PyType_GenericNew;
@@ -1700,9 +2075,4 @@ initgobject(void)
     PyModule_AddObject(m, "TYPE_PARAM", pyg_type_wrapper_new(G_TYPE_PARAM));
     PyModule_AddObject(m, "TYPE_OBJECT", pyg_type_wrapper_new(G_TYPE_OBJECT));
     PyModule_AddObject(m, "TYPE_PYOBJECT", pyg_type_wrapper_new(PY_TYPE_OBJECT));
-
-    if (PyErr_Occurred()) {
-	PyErr_Print();
-	Py_FatalError("can't initialise module gobject");
-    }
 }
