@@ -1,4 +1,24 @@
-/* -*- Mode: C; c-basic-offset: 4 -*- */
+/* -*- Mode: C; c-basic-offset: 4 -*-
+ * pygtk- Python bindings for the GTK toolkit.
+ * Copyright (C) 1998-2003  James Henstridge
+ *
+ *   gtk-types.c: wrappers for some specialised GTK types.
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
+ * USA
+ */
 #include <gtk/gtk.h>
 #include "pygtk-private.h"
 #include <structmember.h>
@@ -55,7 +75,9 @@ _pygtk_style_helper_new(GtkStyle *style, int type, gpointer array)
 static void
 pygtk_style_helper_dealloc(PyGtkStyleHelper_Object *self)
 {
+    pyg_unblock_threads();
     g_object_unref(self->style);
+    pyg_block_threads();
     PyObject_DEL(self);
 }
 
@@ -126,7 +148,11 @@ pygtk_style_helper_setitem(PyGtkStyleHelper_Object *self, int pos,
 		PyErr_SetString(PyExc_TypeError, "can only assign a GdkGC");
 		return -1;
 	    }
-	    if (array[pos]) g_object_unref(array[pos]);
+	    if (array[pos]) {
+		pyg_unblock_threads();
+		g_object_unref(array[pos]);
+		pyg_block_threads();
+	    }
 	    array[pos] = GDK_GC(g_object_ref(pygobject_get(value)));
 	    return 0;
 	}
@@ -139,7 +165,11 @@ pygtk_style_helper_setitem(PyGtkStyleHelper_Object *self, int pos,
 				"can only assign a GdkPixmap or None");
 		return -1;
 	    }
-	    if (array[pos]) g_object_unref(array[pos]);
+	    if (array[pos]) {
+		pyg_unblock_threads();
+		g_object_unref(array[pos]);
+		pyg_block_threads();
+	    }
 	    if (value != Py_None)
 		array[pos] = GDK_PIXMAP(g_object_ref(pygobject_get(value)));
 	    else
@@ -790,7 +820,9 @@ _pygtk_tree_model_row_new(GtkTreeModel *model, GtkTreeIter *iter)
 static void
 pygtk_tree_model_row_dealloc(PyGtkTreeModelRow *self)
 {
+    pyg_unblock_threads();
     g_object_unref(self->model);
+    pyg_block_threads();
     PyObject_DEL(self);
 }
 
@@ -940,7 +972,7 @@ static PyGetSetDef pygtk_tree_model_row_getsets[] = {
 static PyTypeObject PyGtkTreeModelRow_Type = {
     PyObject_HEAD_INIT(NULL)
     0,
-    "gtk.gtk.TreeModelRow",
+    "gtk.TreeModelRow",
     sizeof(PyGtkTreeModelRow),
     0,
     (destructor)pygtk_tree_model_row_dealloc,
@@ -998,7 +1030,9 @@ _pygtk_tree_model_row_iter_new(GtkTreeModel *model, GtkTreeIter *parent_iter)
 static void
 pygtk_tree_model_row_iter_dealloc(PyGtkTreeModelRowIter *self)
 {
+    pyg_unblock_threads();
     g_object_unref(self->model);
+    pyg_block_threads();
     PyObject_DEL(self);
 }
 
@@ -1030,7 +1064,7 @@ pygtk_tree_model_row_iter_next(PyGtkTreeModelRowIter *self)
 static PyTypeObject PyGtkTreeModelRowIter_Type = {
     PyObject_HEAD_INIT(NULL)
     0,
-    "gtk.gtk.TreeModelRowIter",
+    "gtk.TreeModelRowIter",
     sizeof(PyGtkTreeModelRowIter),
     0,
     (destructor)pygtk_tree_model_row_iter_dealloc,
@@ -1058,6 +1092,53 @@ static PyTypeObject PyGtkTreeModelRowIter_Type = {
     (iternextfunc)pygtk_tree_model_row_iter_next
 };
 
+int
+_pygtk_tree_model_set_row(GtkTreeModel *model, GtkTreeIter *iter,
+			  PyObject *items)
+{
+    gint n_columns, i;
+
+    if (!GTK_IS_LIST_STORE(model) && !GTK_IS_TREE_STORE(model)) {
+	PyErr_SetString(PyExc_TypeError,
+			"can not set cells in this tree model");
+	return -1;
+    }
+
+    if (!PySequence_Check(items)) {
+	PyErr_SetString(PyExc_TypeError, "expecting a sequence");
+	return -1;
+    }
+    n_columns = gtk_tree_model_get_n_columns(model);
+    if (PySequence_Length(items) != n_columns) {
+	PyErr_SetString(PyExc_ValueError, "row sequence has wrong length");
+	return -1;
+    }
+    for (i = 0; i < n_columns; i++) {
+	GValue value = { 0, };
+	PyObject *item;
+
+	item = PySequence_GetItem(items, i);
+	if (!item)
+	    return -1;
+	g_value_init(&value, gtk_tree_model_get_column_type(model, i));
+	if (pyg_value_from_pyobject(&value, item)) {
+	    Py_DECREF(item);
+	    PyErr_SetString(PyExc_TypeError,
+			    "value is of wrong type for this column");
+	    return -1;
+	}
+
+	if (GTK_IS_LIST_STORE(model))
+	    gtk_list_store_set_value(GTK_LIST_STORE(model), iter, i, &value);
+	else if (GTK_IS_TREE_STORE(model))
+	    gtk_tree_store_set_value(GTK_TREE_STORE(model), iter, i, &value);
+
+	g_value_unset(&value);
+	Py_DECREF(item);
+    }
+    return 0;
+}
+
 PyObject *
 pygtk_tree_path_to_pyobject(GtkTreePath *path)
 {
@@ -1075,7 +1156,12 @@ pygtk_tree_path_to_pyobject(GtkTreePath *path)
 GtkTreePath *
 pygtk_tree_path_from_pyobject(PyObject *object)
 {
-    if (PyInt_Check(object)) {
+    if (PyString_Check(object)) {
+	GtkTreePath *path;
+
+	path = gtk_tree_path_new_from_string(PyString_AsString(object));
+	return path;
+    } else if (PyInt_Check(object)) {
 	GtkTreePath *path;
 
 	path = gtk_tree_path_new();
